!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!  Sparse Cholesky factorisation and inversion of a positive definite matrix
!              using a block-wise ("super-nodal") approach
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

! Computations require BLAS and LAPACK libraries

! Asssume: Matrix is stored in compressed format (see George and Liu, 1981) & 
!          re-ordering to reduce fill-in and symbolic factorisation have 
!          been carried out

! 1. Routine "super_nodes" works out the node boundaries and number of nodes,
!    starting from the last row

! 2. Routine "super_gsfct" performs the Cholesky factorisation;
!    it replaces George & Liu's routine GSFCT, but uses a right- rather than
!    left-looking algorithm

! 3. Routine "super_sparsinv" takes the Cholesky factor as input and returns
!    the *subset* of elements of the sparse inverse corresponding to non-zero
!    elements in the Cholesky factor (Only those! This means that the product 
!    of the `inverse' obtained & the original matrix usually is not quite an 
!    identity matrix)

! Parameters:
! neqns   = no. of rows/columns 
! ixvec1  = index vector: elements in column i of the lower triangle of the 
!           matrix are held in position ixvec1(i) to ixvec1(i+1)-1 of xspars
!           (equivalent to IXLNZ in GSFCT)
! ixvec2  = adjacency structure -> starting position
!           (equivalent to XNZSUB in GSFCT)
! ixsub   = adjacency structure -> column no.s of non-zero entries for each row
!           (equivalent to NZSUB in GSFCT)
! dia     = diagonal elements of matrix/ cholesky fcator/ inverse
! xspars  = non-zero off-diagonal elements (lower triangle) of  matrix/ 
!           cholesky factor/ inverse
!           (equivalent to LNZ in GSFCT)
! nnode   = no. of supernodes
! inode   = last row/column in each node; backwards

!                                                             @C K. Meyer 2013
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

! ============================================================================
  subroutine super_nodes( neqns, ixvec1, ixvec2, ixsub, nnode, inode )
! ============================================================================

  implicit none
  integer, intent(in)                      :: neqns
  integer, dimension(*), intent(in)        :: ixsub ,ixvec1, ixvec2
  integer, intent(out)                     :: nnode
  integer, dimension(neqns), intent(out)   :: inode
  integer                                  :: i, ii, j, n, ilast, kk
  real(8)                                  :: xx

! establish boundaries between diagonal blocks 100% full
  ilast = neqns; nnode = 0
  do i = neqns-1, 1, -1
     ii = ixvec2(i); n = 0
     do j = ixvec1(i), ixvec1(i+1)-1
        kk = ixsub(ii); ii = ii + 1
        if(  kk > ilast ) exit
        n = n + 1
     end do
     xx = dble( n ) / dble( ilast - i )
     if( xx < 1.d0 ) then
         nnode = nnode +1; inode(nnode) = ilast; ilast = i
     end if
  end do
  nnode = nnode +1; inode(nnode) = ilast; inode(nnode+1) = 0
  write(*,*) 'no. of rows =',neqns,'  no. of super-nodes =',nnode

  return
  end subroutine super_nodes

! ============================================================================
  subroutine super_gsfct( neqns, ixvec1, xspars, ixvec2, ixsub, dia, nnode,   &
&                                                                    inode )
! ============================================================================

  implicit none
  integer, intent(in)                      :: neqns, nnode
  real(8), intent(inout), dimension(neqns) :: dia
  real(8), dimension(*), intent(inout)     :: xspars
  integer, dimension(*), intent(in)        :: ixsub ,ixvec1, ixvec2, inode
  integer, dimension(:), allocatable       :: jvec, kvec
  real(8), dimension(:,:), allocatable     :: ttt, s21, s22
  integer                                  :: i, j, k, jrow, n, ksub, irow,    &
&                                             jnode, icol1, icol2, jcol, ii,   &
&                                             jj, mm, kk

  allocate( jvec(neqns), kvec(neqns),stat = ii ); if( ii /= 0 ) call alloc_err

  do jnode = nnode, 1, -1
     icol1 = inode(jnode+1) + 1; icol2 = inode(jnode);  mm = icol2 - icol1 +1

!    pick out diagonal block
     allocate( ttt(icol1:icol2,icol1:icol2), stat = ii )
     if( ii /= 0 ) call alloc_err
     ttt = 0.d0; jvec = 0
     do irow = icol1, icol2
        ttt(irow,irow) = dia(irow)
        ksub = ixvec2(irow)
        do i = ixvec1(irow), ixvec1(irow+1)-1
           jcol = ixsub(ksub); ksub = ksub + 1
           if( jcol <= icol2 ) then
               ttt(jcol,irow) = xspars(i) 
           else
               jvec(jcol) = 1
           end if
        end do
      end do

!     factorise
      call dpotrf( 'L', mm, ttt, mm, ii )
      if( ii /= 0 ) then
          write(*,*)'Dense fact: Routine DPOTRF returned error code', ii
          write(*,*)'... coefficient matrix must be positive definite'; stop
      end if

!     adjust block below diagonal
!     .. count no. of rows
      n = 0
      do i = icol2+1, neqns
         if( jvec(i) == 0 ) cycle
         n = n + 1; jvec(i) = n; kvec(n) = i
      end do
      if( n > 0 ) then
!         ... pick out rows
          allocate( s21(n, icol1:icol2), stat = ii )
          if( ii /= 0 ) call alloc_err
          s21 = 0.d0
          do irow = icol1, icol2
             ksub = ixvec2(irow)
             do i = ixvec1(irow), ixvec1(irow+1)-1
                jcol = ixsub(ksub); ksub = ksub + 1
                if( jcol <= icol2 ) cycle
                jj = jvec(jcol); s21(jj,irow) = xspars(i)
             end do
          end do

!         calculate L21
          call dtrsm( 'R', 'L', 'T', 'N', n, mm, 1.d0, ttt, mm, s21, n )

!         adjust remaining triangle to right: A22 := A22 - L21 L21'
          allocate( s22(n,n), stat = ii ); if( ii /= 0 ) call alloc_err
          call dsyrk( 'L', 'N', n, mm, 1.d0, s21, n, 0.d0, s22, n )
          kk = kvec(n)
          do i = 1, n
              jrow = kvec(i); dia(jrow) = dia(jrow) - s22(i,i)
              ksub = ixvec2(jrow)
              do j = ixvec1(jrow), ixvec1(jrow+1)-1
                 jcol = ixsub(ksub); if( jcol > kk ) exit
                 ksub = ksub + 1; ii = jvec(jcol)
                 if( ii > 0 ) xspars(j) = xspars(j) - s22(ii,i)
              end do
          end do
          deallocate( s22 )
      end if

!     transfer block back to sparse storage
      do irow = icol1, icol2 
         dia(irow) = ttt(irow,irow)
         ksub = ixvec2(irow)
         do i = ixvec1(irow), ixvec1(irow+1)-1
            jcol = ixsub(ksub); ksub = ksub + 1
            if( jcol <= icol2 ) then
                xspars(i) = ttt(jcol,irow)
            else
                xspars(i) = s21( jvec(jcol), irow)
            end if
         end do
      end do
      deallocate( ttt ); if( n > 0 ) deallocate( s21)

  end do ! jnode

  deallocate( jvec, kvec )

  return 
  end subroutine super_gsfct

! ============================================================================
  subroutine super_sparsinv( neqns, ixvec1, xspars, ixvec2, ixsub, dia, nnode, &
&                                                                       inode )
! ============================================================================

  implicit none
  integer, intent(in)                      :: neqns, nnode
  real(8), intent(inout), dimension(neqns) :: dia
  real(8), dimension(*), intent(inout)     :: xspars
  integer, dimension(*), intent(in)        :: ixsub ,ixvec1, ixvec2, inode
  real(8), dimension(:,:), allocatable     :: ttt, s21, s22, f21
  real(8), dimension(:), allocatable       :: rr, qx
  integer, dimension(:), allocatable       :: kvec, jvec
  real(8)                                  :: tt,  xx
  integer                                  :: irow, ksub, i, j, k, m, jcol,    &
&                                             jrow, jnode, icol2, icol1, ii,   &
&                                             jj, mm, n21, iopt

  allocate( jvec(neqns), kvec(neqns),stat = ii ); if( ii /= 0 ) call alloc_err

! backwards flops: determine inverse using supernodal blocks
  do jnode = 1, nnode
     icol1 = inode(jnode+1) + 1; icol2 = inode(jnode);  mm = icol2 - icol1 +1
      
!    pick out diagonal block
     allocate( ttt(icol1:icol2,icol1:icol2), stat = ii )
     if( ii /= 0 ) call alloc_err
     ttt = 0.d0; jvec = 0
     do irow = icol1, icol2
        ttt(irow,irow) = dia(irow);  ksub = ixvec2(irow)
        do i = ixvec1(irow), ixvec1(irow+1)-1
           jcol = ixsub(ksub); ksub = ksub + 1
           if( jcol <= icol2 ) then
               ttt(jcol,irow) = xspars(i) 
          else 
               jvec(jcol) = 1
           end if
        end do
     end do

!    pick out lead columns (condensed)
     n21 = 0
     do i = icol2+1, neqns
        if( jvec(i) == 0 ) cycle
        n21 = n21 + 1; jvec(i) = n21; kvec(n21) = i
     end do
     if( n21 > 0 ) then
         allocate( s21(n21, icol1:icol2), stat = ii )
         if( ii /= 0 ) call alloc_err
         allocate( f21(n21, icol1:icol2), stat = ii )
         if( ii /= 0 ) call alloc_err
         s21 = 0.d0; f21 = 0.d0
         do irow = icol1, icol2
            ksub = ixvec2(irow)
            do i = ixvec1(irow), ixvec1(irow+1)-1
               jcol = ixsub(ksub); ksub = ksub + 1
               if( jcol <= icol2 ) cycle
               jj = jvec(jcol); s21(jj,irow) = xspars(i)
            end do
         end do
!        ... post-multiply with inverse Chol factor -> solve
         call dtrsm( 'R', 'L', 'N', 'N', n21, mm, 1.d0, ttt, mm, s21, n21 )
!        ... invert Cholesky factor
         call dpotri( 'L',  mm, ttt, mm, ii )
         if( ii /= 0 ) then
             write(*,*) 'Routine DPOTRI returned error code', ii; stop
         end if
!        ... pre-multiply by already inverted submatrix
         iopt = 2
44       if( iopt == 1 ) then
            allocate( rr(icol1:icol2), qx(icol1:icol2), stat = ii )
            if( ii /= 0 ) call alloc_err
            f21 = 0.d0
            do k = 1, n21
               jrow = kvec(k); rr = s21(k,:); qx =  dia(jrow) * rr 
               ksub = ixvec2(jrow)
               do i = ixvec1(jrow), ixvec1(jrow+1)-1
                  jcol = ixsub(ksub); ksub = ksub + 1
                  m = jvec(jcol); if( m < 1 ) cycle
                  xx = xspars(i); f21(m,:) = f21(m,:) + xx * rr
                  qx = qx + xx * s21(m,:) 
               end do
               f21(k,:) = f21(k,:) + qx 
            end do
            deallocate( rr, qx )
         else
            allocate( s22(n21,n21), stat = ii )
            if( ii /= 0 ) then
                iopt = 1; go to 44
            end if
            s22 = 0.d0
            do k = 1, n21
               jrow = kvec(k); s22(k,k) = dia(jrow)
               ksub = ixvec2(jrow)
               do i = ixvec1(jrow), ixvec1(jrow+1)-1
                  jcol = ixsub(ksub); ksub = ksub + 1
                  m = jvec(jcol); if( m > 0 ) s22(m,k) = xspars(i)
               end do
            end do
            call dsymm( 'L', 'L', n21, mm, 1.d0, s22, n21, s21, n21, 0.d0,     &
&                                                                f21, n21  )
            deallocate( s22 )
         end if
!        ... adjustments to current block
         call dgemm( 'T', 'N', mm, mm, n21, 1.d0, f21, n21, s21, n21, 1.d0,    &
&                                                                 ttt, mm )
     else
         call dpotri( 'L',  mm, ttt, mm, ii )
         if( ii /= 0 ) then
             write(*,*) 'Routine DPOTRI returned error code', ii; stop
         end if
     end if

!    save current block  
     do irow = icol1, icol2
        dia(irow) = ttt(irow,irow); ksub = ixvec2(irow)
        do i = ixvec1(irow), ixvec1(irow+1)-1
           jcol = ixsub(ksub); ksub = ksub + 1
           if( jcol <= icol2 ) then
               xspars(i) = ttt(jcol,irow) 
           else 
               xspars(i) = - f21(jvec(jcol),irow)
           end if
        end do
     end do
     deallocate( ttt ); if( n21 > 0 ) deallocate( s21, f21 )
  end do ! jnode

  deallocate( jvec, kvec )

  return
  end subroutine super_sparsinv

! ====================
  subroutine alloc_err
! ====================

  write(*,*) &
& 'Super-nodal factorisation/inversion: Failed to allocate work space required'

  stop
  end subroutine alloc_err

